from .Adv_Loss import Adv_Loss
from .ComplEx_NNE_AER_Loss import ComplEx_NNE_AER_Loss
from .SimplE_Loss import SimplE_Loss
from .Cross_Entropy_Loss import Cross_Entropy_Loss
from .RGCN_Loss import RGCN_Loss
from .KBAT_Loss import KBAT_Loss
from .CrossE_Loss import CrossE_Loss
from .IterE_Loss import IterE_Loss
from .Margin_Loss import Margin_Loss
from .RugE_Loss import RugE_Loss